@@ -7,19 +7,30 @@
  * Author: Jaka Mocnik  <jaka@gnu.org>
  */
 
-#include <config.h>
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif /* HAVE_CONFIG_H */
 
 #include <gnome.h>
+
 #include <bonobo-activation/bonobo-activation.h>
+
 #include <libgnomeui/gnome-window-icon.h>
+
 #include <bonobo.h>
 #include <bonobo/bonobo-ui-main.h>
+
 #include <libgnomevfs/gnome-vfs-ops.h>
+
 #include <math.h>
 #include <ctype.h>
 
 #include "ghex-window.h"
-#include "ghex.h"
+#include "ui.h"
+#include "converter.h"
+#include "chartable.h"
+#include "configuration.h"
+#include "hex-dialog.h"
 
 #define GHEX_WINDOW_DEFAULT_WIDTH 320
 #define GHEX_WINDOW_DEFAULT_HEIGHT 256
@@ -70,16 +81,42 @@ ghex_window_drag_data_received(GtkWidget *widget,
         else
             newwin = NULL;
         while(*uri) {
-            if(!g_ascii_strncasecmp("file:", *uri, 5)) {
-                if(newwin == NULL) {
-                    newwin = ghex_window_new_from_file((*uri) + 5);
+            GError *err = NULL;
+            gchar *filename = g_filename_from_uri(*uri, NULL, &err);
+
+            uri++;
+            if(filename == NULL) {
+                GtkWidget *dlg;
+                dlg = gtk_message_dialog_new(GTK_WINDOW(win),
+                                             GTK_DIALOG_MODAL,
+                                             GTK_MESSAGE_ERROR,
+                                             GTK_BUTTONS_OK,
+                                             _("Can not open URI:\n%s"),
+                                             err->message);
+                g_error_free(err);
+                continue;
+            }
+
+	    	if(newwin == NULL)
+                newwin = ghex_window_new();
+            if(ghex_window_load(GHEX_WINDOW(newwin), filename)) {
+                if(newwin != GTK_WIDGET(win))
                     gtk_widget_show(newwin);
-                }
-                else
-                    ghex_window_load(GHEX_WINDOW(newwin), (*uri) + 5);
                 newwin = NULL;
             }
-            uri++;
+            else {
+                GtkWidget *dlg;
+                dlg = gtk_message_dialog_new(GTK_WINDOW(win),
+                                             GTK_DIALOG_MODAL,
+                                             GTK_MESSAGE_ERROR,
+                                             GTK_BUTTONS_OK,
+                                             _("Can not open file:\n%s"),
+                                             filename);
+                gtk_widget_show(dlg);
+                gtk_dialog_run(GTK_DIALOG(dlg));
+                gtk_widget_destroy(dlg);
+            }
+            g_free(filename);
         }
         g_strfreev(win->uris_to_open);
         win->uris_to_open = NULL;
@@ -591,6 +628,39 @@ ghex_window_sync_group_type(GHexWindow *win)
     bonobo_ui_component_set_prop(win->uic, group_path, "state", "1", NULL);
 }
 
+void
+ghex_window_update_status_message(GHexWindow *win)
+{
+#define FMT_LEN    128
+#define STATUS_LEN 128
+
+    gchar fmt[FMT_LEN], status[STATUS_LEN];
+    gint current_pos;
+    gint ss, se, len;
+
+    if(NULL == win->gh) {
+        ghex_window_show_status(win, " ");
+        return;
+    }
+
+    current_pos = gtk_hex_get_cursor(win->gh);
+    if(g_snprintf(fmt, FMT_LEN, _("Offset: %s"), offset_fmt) < FMT_LEN) {
+        g_snprintf(status, STATUS_LEN, fmt, current_pos);
+        if(gtk_hex_get_selection(win->gh, &ss, &se)) {
+            if(g_snprintf(fmt, FMT_LEN, _("; %s bytes from %s to %s selected"),
+                          offset_fmt, offset_fmt, offset_fmt) < FMT_LEN) {
+                len = strlen(status);
+                if(len < STATUS_LEN) {
+                    g_snprintf(status + len, STATUS_LEN - len, fmt, se - ss, ss, se);
+                }
+            }
+        }
+        ghex_window_show_status(win, status);
+    }
+    else
+        ghex_window_show_status(win, " ");
+}
+
 static void
 cursor_moved_cb(GtkHex *gtkhex, gpointer user_data)
 {
@@ -601,17 +671,10 @@ cursor_moved_cb(GtkHex *gtkhex, gpointer user_data)
 	GHexWindow *win = GHEX_WINDOW(user_data);
 
     current_pos = gtk_hex_get_cursor(gtkhex);
-
-	if((format = g_strdup_printf(_("Offset: %s"), offset_fmt)) != NULL) {
-		if((cursor_pos = g_strdup_printf(format, current_pos)) != NULL) {
-			ghex_window_show_status(win, cursor_pos);
-            g_free(cursor_pos);
-		}
-		g_free(format);
-	}
+    ghex_window_update_status_message(win);
     for (i = 0; i < 8; i++)
     {
-        /* returns  0 on buffer overflow, which is what we want */
+        /* returns 0 on buffer overflow, which is what we want */
         val.v[i] = gtk_hex_get_byte(gtkhex, current_pos+i);
     }
     hex_dialog_updateview(win->dialog, &val);
@@ -641,7 +704,7 @@ ghex_window_load(GHexWindow *win, const gchar *filename)
     else
         full_path = g_strdup(filename);
     
-    doc = hex_document_new(full_path);
+    doc = hex_document_new_from_file(full_path);
     g_free(full_path);
     if(!doc)
         return FALSE;
@@ -650,17 +713,24 @@ ghex_window_load(GHexWindow *win, const gchar *filename)
     gtk_hex_show_offsets(GTK_HEX(gh), show_offsets_column);
     g_signal_connect(G_OBJECT(doc), "document_changed",
                      G_CALLBACK(ghex_window_doc_changed), win);
+    g_signal_connect(G_OBJECT(doc), "undo",
+                     G_CALLBACK(set_doc_menu_sensitivity), win);;
+    g_signal_connect(G_OBJECT(doc), "redo",
+                     G_CALLBACK(set_doc_menu_sensitivity), win);;
+    g_signal_connect(G_OBJECT(doc), "undo_stack_forget",
+                     G_CALLBACK(set_doc_menu_sensitivity), win);;
     g_signal_connect(G_OBJECT(gh), "cursor_moved",
                      G_CALLBACK(cursor_moved_cb), win);
     gtk_widget_show(gh);
 
     vbox = gtk_vbox_new(FALSE, 0);
+    gtk_container_set_border_width(GTK_CONTAINER(win), GNOME_PAD_SMALL);
     gtk_widget_show(vbox);
-    gtk_box_pack_start(GTK_BOX(vbox), gh, TRUE, TRUE, GNOME_PAD);
+    gtk_box_pack_start(GTK_BOX(vbox), gh, TRUE, TRUE, GNOME_PAD_SMALL);
 
     win->dialog = hex_dialog_new();
     win->dialog_widget = hex_dialog_getview(win->dialog);
-    gtk_box_pack_start(GTK_BOX(vbox), win->dialog_widget, FALSE, FALSE, GNOME_PAD);
+    gtk_box_pack_start(GTK_BOX(vbox), win->dialog_widget, FALSE, FALSE, GNOME_PAD_SMALL);
     state = bonobo_ui_component_get_prop (win->uic, "/commands/TypeDialog", "state", NULL);
     if ((state && atoi(state)) || !state)
     {
@@ -870,7 +940,7 @@ void
 ghex_window_set_doc_name(GHexWindow *win, const gchar *name)
 {
     if(name != NULL) {
-        gchar *title = g_strdup_printf(_("GHex - %s"), name);
+        gchar *title = g_strdup_printf(_("%s - GHex"), name);
         gtk_window_set_title(GTK_WINDOW(win), title);
         g_free(title);
     }
@@ -901,7 +971,7 @@ remove_message_timeout (MessageInfo * mi)
 
 	/* Remove the status message */
 	/* NOTE : Use space ' ' not an empty string '' */
-	ghex_window_show_status (mi->win, " ");
+	ghex_window_update_status_message (mi->win);
     g_signal_handlers_disconnect_by_func(G_OBJECT(mi->win),
                                          remove_timeout_cb, mi);
 	g_free (mi);
@@ -952,7 +1022,6 @@ ghex_window_flash (GHexWindow * win, const gchar * flash)
 	g_return_if_fail (GHEX_IS_WINDOW (win));
 	g_return_if_fail (flash != NULL);
 
-
 	mi = g_new (MessageInfo, 1);
 
 	ghex_window_show_status (win, flash);
@@ -1077,7 +1146,7 @@ ghex_window_save_as(GHexWindow *win)
                     doc->file_name = g_strdup(filename);
                     doc->changed = FALSE;
                     win->changed = FALSE;
-                    
+
                     for(i = strlen(doc->file_name);
                         (i >= 0) && (doc->file_name[i] != '/');
                         i--)
@@ -1174,7 +1243,7 @@ ghex_window_ok_to_close(GHexWindow *win)
                 return FALSE;
             }
             else if(!hex_document_write(doc)) {
-                display_error_dialog(win, _("An error occured while saving file!"));
+                display_error_dialog(win, _("An error occurred while saving file!"));
 				return FALSE;
 			}
 		}
