@@ -21,13 +21,17 @@
    Author: Jaka Mocnik <jaka@gnu.org>
  */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif /* HAVE_CONFIG_H */
+
 #include <gnome.h>
 #include <gdk/gdkkeysyms.h>
 
-#include <hex-document.h>
-#include <gtkhex.h>
-
-#include <ghex-marshal.h>
+#include "hex-document.h"
+#include "gtkhex.h"
+#include "gtkhex-private.h"
+#include "ghex-marshal.h"
 
 #define DISPLAY_BORDER 4
 
@@ -36,12 +40,9 @@
 
 #define SCROLL_TIMEOUT 100
 
-/* jaKa: this causes trouble to pango */
-#if 0 
-#  define is_displayable(c) ((char_widths[(guchar)c])?1:0)
-#else 
-#  define is_displayable(c) (((c) >= 0x20) && ((c) < 0x7f))
-#endif /* 0 */
+#define DEFAULT_FONT "Monospace 12"
+
+#define is_displayable(c) (((c) >= 0x20) && ((c) < 0x7f))
 
 typedef void (*DataChangedSignal)(GtkObject *, gpointer, gpointer);
 
@@ -82,8 +83,6 @@ static void render_hex_highlights(GtkHex *gh, gint cursor_line);
 static void render_ascii_highlights(GtkHex *gh, gint cursor_line);
 static void render_hex_lines(GtkHex *, gint, gint);
 static void render_ascii_lines(GtkHex *, gint, gint);
-static void gtk_hex_set_selection(GtkHex *gh, gint start, gint end);
-
 
 static void gtk_hex_validate_highlight(GtkHex *gh, GtkHex_Highlight *hl);
 static void gtk_hex_invalidate_highlight(GtkHex *gh, GtkHex_Highlight *hl);
@@ -803,16 +802,18 @@ static void recalc_displays(GtkHex *gh, guint width, guint height) {
 	if(gh->cpl == 0)
 		return;
 
-	gh->lines = gh->document->file_size / gh->cpl;
-	if(gh->document->file_size % gh->cpl)
-		gh->lines++;
-	
+	if(gh->document->file_size == 0)
+		gh->lines = 1;
+	else {
+		gh->lines = gh->document->file_size / gh->cpl;
+		if(gh->document->file_size % gh->cpl)
+			gh->lines++;
+	}
+
 	gh->vis_lines = ( (gint) (height - 2*GTK_CONTAINER(gh)->border_width - 2*widget_get_yt(GTK_WIDGET(gh))) ) / ( (gint) gh->char_height );
 
 	gh->adisp_width = gh->cpl*gh->char_width + 1;
-	xcpl = gh->cpl*2 + gh->cpl/gh->group_type;
-	if(gh->cpl % gh->group_type == 0)
-		xcpl--;
+	xcpl = gh->cpl*2 + (gh->cpl - 1)/gh->group_type;
 	gh->xdisp_width = xcpl*gh->char_width + 1;
 
 	if(gh->disp_buffer)
@@ -936,6 +937,10 @@ static gboolean scroll_timeout_handler(GtkHex *gh) {
 	return TRUE;
 }
 
+static void hex_scroll_cb(GtkWidget *w, GdkEventScroll *event, GtkHex *gh) {
+	gtk_widget_event(gh->scrollbar, (GdkEvent *)event);
+}
+
 static void hex_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 	if( (event->type == GDK_BUTTON_RELEASE) &&
 		(event->button == 1) ) {
@@ -986,7 +991,7 @@ static void hex_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 		}
 		gh->button = 0;
 	}
-	else 
+	else
 		gh->button = 0;
 }
 
@@ -1024,6 +1029,10 @@ static void hex_motion_cb(GtkWidget *w, GdkEventMotion *event, GtkHex *gh) {
 	}
 }
 
+static void ascii_scroll_cb(GtkWidget *w, GdkEventScroll *event, GtkHex *gh) {
+	gtk_widget_event(gh->scrollbar, (GdkEvent *)event);
+}
+
 static void ascii_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 	if( (event->type == GDK_BUTTON_RELEASE) &&
 		(event->button == 1) ) {
@@ -1229,7 +1238,7 @@ static void primary_get_cb(GtkClipboard *clipboard,
 		end_pos = MAX(gh->selection.start, gh->selection.end);
  
 		text = hex_document_get_data(gh->document, start_pos,
-											 end_pos - start_pos);
+									 end_pos - start_pos);
 		gtk_selection_data_set_text(data, text, end_pos - start_pos);
 		g_free(text);
 	}
@@ -1239,7 +1248,7 @@ static void primary_clear_cb(GtkClipboard *clipboard,
 							 gpointer user_data_or_owner) {
 }
 
-static void gtk_hex_set_selection(GtkHex *gh, gint start, gint end)
+void gtk_hex_set_selection(GtkHex *gh, gint start, gint end)
 {
 	gint length = gh->document->file_size;
 	gint oe, os, ne, ns;
@@ -1283,11 +1292,30 @@ static void gtk_hex_set_selection(GtkHex *gh, gint start, gint end)
 									gh);
 }
 
-void gtk_hex_select_region (GtkHex *gh,
-							gint         start,
-							gint         end) {
-	
-	gtk_hex_set_selection(gh, start, end);
+gboolean gtk_hex_get_selection(GtkHex *gh, gint *start, gint *end)
+{
+	gint ss, se;
+
+	if(gh->selection.start > gh->selection.end) {
+		se = gh->selection.start;
+		ss = gh->selection.end;
+	}
+	else {
+		ss = gh->selection.start;
+		se = gh->selection.end;
+	}
+
+	if(NULL != start)
+		*start = ss;
+	if(NULL != end)
+		*end = se;
+
+	return !(ss == se);
+}
+
+void gtk_hex_clear_selection(GtkHex *gh)
+{
+	gtk_hex_set_selection(gh, 0, 0);
 }
 
 void gtk_hex_delete_selection(GtkHex *gh)
@@ -1298,7 +1326,7 @@ void gtk_hex_delete_selection(GtkHex *gh)
 	start = MIN(gh->selection.start, gh->selection.end);
 	end = MAX(gh->selection.start, gh->selection.end);
 
-	gtk_hex_select_region(gh, 0, 0);
+	gtk_hex_set_selection(gh, 0, 0);
 
 	if(start != end) {
 		if(start < gh->cursor_pos)
@@ -1490,22 +1518,22 @@ static void gtk_hex_update_all_auto_highlights(GtkHex *gh, gboolean delete, gboo
 	}
 }
 
-void gtk_hex_copy_clipboard(GtkHex *gh)
+void gtk_hex_copy_to_clipboard(GtkHex *gh)
 {
 	g_signal_emit_by_name(G_OBJECT(gh), "copy_clipboard");
 }
 
-void gtk_hex_cut_clipboard(GtkHex *gh)
+void gtk_hex_cut_to_clipboard(GtkHex *gh)
 {
 	g_signal_emit_by_name(G_OBJECT(gh), "cut_clipboard");
 }
 
-void gtk_hex_paste_clipboard(GtkHex *gh)
+void gtk_hex_paste_from_clipboard(GtkHex *gh)
 {
 	g_signal_emit_by_name(G_OBJECT(gh), "paste_clipboard");
 }
 
-static void gtk_hex_real_copy_clipboard(GtkHex *gh)
+static void gtk_hex_real_copy_to_clipboard(GtkHex *gh)
 {
 	gint start_pos; 
 	gint end_pos;
@@ -1522,15 +1550,15 @@ static void gtk_hex_real_copy_clipboard(GtkHex *gh)
 	}
 }
 
-static void gtk_hex_real_cut_clipboard(GtkHex *gh)
+static void gtk_hex_real_cut_to_clipboard(GtkHex *gh)
 {
 	if(gh->selection.start != -1 && gh->selection.end != -1) {
-		gtk_hex_real_copy_clipboard(gh);
+		gtk_hex_real_copy_to_clipboard(gh);
 		gtk_hex_delete_selection(gh);
 	}
 }
 
-static void gtk_hex_real_paste_clipboard(GtkHex *gh)
+static void gtk_hex_real_paste_from_clipboard(GtkHex *gh)
 {
 	GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
 	gchar *text;
@@ -1850,9 +1878,9 @@ static void gtk_hex_class_init(GtkHexClass *klass, gpointer data) {
 	
 	klass->cursor_moved = NULL;
 	klass->data_changed = gtk_hex_real_data_changed;
-	klass->cut_clipboard = gtk_hex_real_cut_clipboard;
-	klass->copy_clipboard = gtk_hex_real_copy_clipboard;
-	klass->paste_clipboard = gtk_hex_real_paste_clipboard;
+	klass->cut_clipboard = gtk_hex_real_cut_to_clipboard;
+	klass->copy_clipboard = gtk_hex_real_copy_to_clipboard;
+	klass->paste_clipboard = gtk_hex_real_paste_from_clipboard;
 
 	klass->primary = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
 	klass->clipboard = gtk_clipboard_get(GDK_NONE);
@@ -1917,7 +1945,7 @@ static void gtk_hex_init(GtkHex *gh, gpointer klass) {
 	gh->xlayout = gtk_widget_create_pango_layout (gh->xdisp, "");
 
 	gtk_widget_set_events (gh->xdisp, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
-						   GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_MOTION_MASK);
+						   GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_MOTION_MASK | GDK_SCROLL_MASK);
 	g_signal_connect(G_OBJECT(gh->xdisp), "expose_event",
 					 G_CALLBACK(hex_expose), gh);
 	g_signal_connect(G_OBJECT(gh->xdisp), "button_press_event",
@@ -1928,6 +1956,8 @@ static void gtk_hex_init(GtkHex *gh, gpointer klass) {
 					 G_CALLBACK(hex_motion_cb), gh);
 	g_signal_connect(G_OBJECT(gh->xdisp), "realize",
 					 G_CALLBACK(hex_realize), gh);
+	g_signal_connect(G_OBJECT(gh->xdisp), "scroll_event",
+					 G_CALLBACK(hex_scroll_cb), gh);
 	gtk_fixed_put(GTK_FIXED(gh), gh->xdisp, 0, 0);
 	gtk_widget_show(gh->xdisp);
 	
@@ -1940,7 +1970,7 @@ static void gtk_hex_init(GtkHex *gh, gpointer klass) {
 	gh->alayout = gtk_widget_create_pango_layout (gh->adisp, "");
 
 	gtk_widget_set_events (gh->adisp, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
-						   GDK_BUTTON_RELEASE_MASK |GDK_BUTTON_MOTION_MASK);
+						   GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_MOTION_MASK | GDK_SCROLL_MASK);
 	g_signal_connect(G_OBJECT(gh->adisp), "expose_event",
 					 G_CALLBACK(ascii_expose), gh);
 	g_signal_connect(G_OBJECT(gh->adisp), "button_press_event",
@@ -1951,6 +1981,8 @@ static void gtk_hex_init(GtkHex *gh, gpointer klass) {
 					 G_CALLBACK(ascii_motion_cb), gh);
 	g_signal_connect(G_OBJECT(gh->adisp), "realize",
 					 G_CALLBACK(ascii_realize), gh);
+	g_signal_connect(G_OBJECT(gh->adisp), "scroll_event",
+					 G_CALLBACK(ascii_scroll_cb), gh);
 	gtk_fixed_put(GTK_FIXED(gh), gh->adisp, 0, 0);
 	gtk_widget_show(gh->adisp);
 	
@@ -2041,6 +2073,8 @@ void gtk_hex_set_cursor(GtkHex *gh, gint index) {
 		if(!gh->insert && index == gh->document->file_size)
 			index--;
 
+		index == MAX(index, 0);
+
 		hide_cursor(gh);
 		
 		gh->cursor_pos = index;
@@ -2096,6 +2130,8 @@ void gtk_hex_set_cursor_xy(GtkHex *gh, gint x, gint y) {
 		if(!gh->insert && cp == gh->document->file_size)
 			cp--;
 
+		cp = MAX(cp, 0);
+
 		hide_cursor(gh);
 		
 		gh->cursor_pos = cp;
@@ -2222,8 +2258,10 @@ void gtk_hex_set_insert_mode(GtkHex *gh, gboolean insert)
 
 	gh->insert = insert;
 
-	if(gh->cursor_pos >= gh->document->file_size)
-		gh->cursor_pos = gh->document->file_size - 1;
+	if(!gh->insert && gh->cursor_pos > 0) {
+		if(gh->cursor_pos >= gh->document->file_size)
+			gh->cursor_pos = gh->document->file_size - 1;
+	}
 }
 
 PangoFontMetrics* gtk_hex_load_font (const char *font_name)
@@ -2296,6 +2334,28 @@ void gtk_hex_delete_autohighlight(GtkHex *gh, GtkHex_AutoHighlight *ahl)
 	g_free(ahl);
 }
 
+void gtk_hex_set_geometry(GtkHex *gh, gint cpl, gint vis_lines)
+{
+	gint width, height, xcpl;
+	GtkRequisition req;
+
+	gtk_widget_size_request(gh->scrollbar, &req);
+
+	if(cpl <= 0 || vis_lines <= 0)
+		return;
+
+	xcpl = 2*cpl + (cpl - 1)/gh->group_type;
+	width = xcpl*gh->char_width + cpl*gh->char_width;
+	width += 2*GTK_CONTAINER(gh)->border_width + 4*widget_get_xt(GTK_WIDGET(gh)) +
+		req.width;
+	if(gh->show_offsets)
+		width += 2*widget_get_xt(GTK_WIDGET(gh)) + 8*gh->char_width;
+
+	height = vis_lines*gh->char_height;
+	height += 2*GTK_CONTAINER(gh)->border_width + 2*widget_get_yt(GTK_WIDGET(gh));
+
+	gtk_widget_set_size_request(GTK_WIDGET(gh), width, height);
+}
 
 void add_atk_namedesc (GtkWidget *widget, const gchar *name, const gchar *desc)
 {
